
/*
 * bltTreeViewStyle.c --
 *
 *	This module implements styles for treeview widget cells.
 *
 * Copyright 1998-1999 Lucent Technologies, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that the copyright notice and warranty
 * disclaimer appear in supporting documentation, and that the names
 * of Lucent Technologies or any of their entities not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 *
 * Lucent Technologies disclaims all warranties with regard to this
 * software, including all implied warranties of merchantability and
 * fitness.  In no event shall Lucent Technologies be liable for any
 * special, indirect or consequential damages or any damages
 * whatsoever resulting from loss of use, data or profits, whether in
 * an action of contract, negligence or other tortuous action, arising
 * out of or in connection with the use or performance of this
 * software.
 *
 *	The "treeview" widget was created by George A. Howlett.
 */

#include "bltInt.h"

#ifndef NO_TREEVIEW

#include "bltTreeView.h"
#include "bltList.h"
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#define STYLE_TEXT_ENTRY	0
#define STYLE_OPTION_MENU	1
#define STYLE_CHECK_BOX		2

#define TEXT_ENTRY_GAP		1

#define DEF_TEXT_ENTRY_SIDE		"left"
#define DEF_TEXT_ENTRY_KEY		(char *)NULL
#define DEF_TEXT_ENTRY_ICON		(char *)NULL
#define DEF_TEXT_ENTRY_BACKGROUND	(char *)NULL
#define DEF_TEXT_ENTRY_FONT		(char *)NULL
#define DEF_TEXT_ENTRY_HIGHLIGHT_BG_COLOR STD_COLOR_NORMAL_BG
#define DEF_TEXT_ENTRY_HIGHLIGHT_FG_COLOR STD_COLOR_NORMAL_FG
#define DEF_TEXT_ENTRY_NORMAL_BG_COLOR 	(char *)NULL
#define DEF_TEXT_ENTRY_NORMAL_FG_COLOR 	(char *)NULL
#define DEF_TEXT_ENTRY_ACTIVE_BG_COLOR 	STD_COLOR_ACTIVE_BG
#define DEF_TEXT_ENTRY_ACTIVE_FG_COLOR 	STD_COLOR_ACTIVE_FG

static Blt_OptionParseProc ObjToIcon;
static Blt_OptionPrintProc IconToObj;
static Blt_OptionFreeProc FreeIcon;
Blt_CustomOption iconOption =
{
    /* Contains a pointer to the widget that's currently being
     * configured.  This is used in the custom configuration parse
     * routine for icons.  */
    ObjToIcon, IconToObj, FreeIcon, NULL,
};

static StyleCreateProc CreateTextEntry, CreateCheckBox, CreateOptionMenu;
static StyleFreeProc FreeTextEntry, FreeCheckBox, FreeOptionMenu;
static StyleDrawProc DrawTextEntry, DrawCheckBox, DrawOptionMenu;
static StyleConfigProc ConfigTextEntry, ConfigCheckBox, ConfigOptionMenu;
static StyleMeasureProc MeasureTextEntry, MeasureCheckBox, MeasureOptionMenu;

static Blt_ConfigSpec textEntrySpecs[] =
{
    {BLT_CONFIG_COLOR, "-activebackground", "activeBackground", 
	"ActiveBackground", DEF_TEXT_ENTRY_ACTIVE_BG_COLOR, 
	Blt_Offset(TreeViewTextEntry, activeBgColor), 0},
    {BLT_CONFIG_SYNONYM, "-activebg", "activeBackground", 
	(char *)NULL, (char *)NULL, 0, 0},
    {BLT_CONFIG_SYNONYM, "-activefg", "activeFackground", 
	(char *)NULL, (char *)NULL, 0, 0},
    {BLT_CONFIG_COLOR, "-activeforeground", "activeForeground", 
	"ActiveForeground", DEF_TEXT_ENTRY_ACTIVE_FG_COLOR, 
	Blt_Offset(TreeViewTextEntry, activeFgColor), 0},
    {BLT_CONFIG_COLOR, "-background", "background", "Background",
     "blue", /* DEF_TEXT_ENTRY_NORMAL_BG_COLOR,  */
	Blt_Offset(TreeViewTextEntry, bgColor), BLT_CONFIG_NULL_OK},
    {BLT_CONFIG_SYNONYM, "-bg", "background", (char *)NULL, (char *)NULL, 
	0, 0},
    {BLT_CONFIG_SYNONYM, "-fg", "foreground", (char *)NULL, (char *)NULL, 
	0, 0},
    {BLT_CONFIG_FONT, "-font", "font", "Font",
	DEF_TEXT_ENTRY_FONT, Blt_Offset(TreeViewTextEntry, font), 
        BLT_CONFIG_NULL_OK},
    {BLT_CONFIG_COLOR, "-foreground", "foreground", "Foreground",
	DEF_TEXT_ENTRY_NORMAL_FG_COLOR, Blt_Offset(TreeViewTextEntry, fgColor),
	BLT_CONFIG_NULL_OK },
    {BLT_CONFIG_COLOR, "-highlightbackground", "highlightBackground",
	"HighlightBackground", DEF_TEXT_ENTRY_HIGHLIGHT_BG_COLOR, 
        Blt_Offset(TreeViewTextEntry, highlightBgColor), BLT_CONFIG_COLOR_ONLY},
    {BLT_CONFIG_COLOR, "-highlightforeground", "highlightForeground", 
	"HighlightForeground", DEF_TEXT_ENTRY_HIGHLIGHT_FG_COLOR, 
	 Blt_Offset(TreeViewTextEntry, highlightFgColor), 0},
    {BLT_CONFIG_SYNONYM, "-highlightbg", "highlightBackground", 
	(char *)NULL, (char *)NULL, 0, 0},
    {BLT_CONFIG_SYNONYM, "-highlightfg", "highlightForeground", 
	(char *)NULL, (char *)NULL, 0, 0},
    {BLT_CONFIG_CUSTOM, "-icon", "icon", "Icon",
	DEF_TEXT_ENTRY_ICON, Blt_Offset(TreeViewTextEntry, icon), 
	BLT_CONFIG_NULL_OK, &iconOption},
    {BLT_CONFIG_STRING, "-key", "key", "key",
	DEF_TEXT_ENTRY_KEY, Blt_Offset(TreeViewTextEntry, key), 
	BLT_CONFIG_NULL_OK, 0},
    {BLT_CONFIG_SIDE, "-side", "side", "side",
	DEF_TEXT_ENTRY_SIDE, Tk_Offset(TreeViewTextEntry, side),
	BLT_CONFIG_DONT_SET_DEFAULT},
    {BLT_CONFIG_END, (char *)NULL, (char *)NULL, (char *)NULL,
	(char *)NULL, 0, 0}
};

static Blt_ConfigSpec optionMenuSpecs[] =
{
    {BLT_CONFIG_END, (char *)NULL, (char *)NULL, (char *)NULL,
	(char *)NULL, 0, 0}
};

static Blt_ConfigSpec checkBoxSpecs[] =
{
    {BLT_CONFIG_END, (char *)NULL, (char *)NULL, (char *)NULL,
	(char *)NULL, 0, 0}
};


/*
 *----------------------------------------------------------------------
 *
 * ObjToIcon --
 *
 *	Convert the name of an icon into a Tk image.
 *
 * Results:
 *	If the string is successfully converted, TCL_OK is returned.
 *	Otherwise, TCL_ERROR is returned and an error message is left in
 *	interpreter's result field.
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
ObjToIcon(clientData, interp, tkwin, objPtr, widgRec, offset)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Interpreter to send results back to */
    Tk_Window tkwin;		/* Not used. */
    Tcl_Obj *objPtr;		/* Tcl_Obj representing the new value. */
    char *widgRec;
    int offset;
{
    TreeView *tvPtr = clientData;
    TreeViewIcon *iconPtr = (TreeViewIcon *)(widgRec + offset);
    TreeViewIcon icon;

    icon = Blt_TreeViewGetIcon(tvPtr, Tcl_GetString(objPtr));
    if (icon == NULL) {
	return TCL_ERROR;
    }
    *iconPtr = icon;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * IconToObj --
 *
 *	Converts the icon into its string representation (its name).
 *
 * Results:
 *	The name of the icon is returned.
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static Tcl_Obj *
IconToObj(clientData, interp, tkwin, widgRec, offset)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;
    Tk_Window tkwin;		/* Not used. */
    char *widgRec;
    int offset;
{
    TreeViewIcon icon = *(TreeViewIcon *)(widgRec + offset);

    return  Tcl_NewStringObj(Blt_NameOfImage((icon)->tkImage), -1);
}

/*ARGSUSED*/
static void
FreeIcon(clientData, display, widgRec, offset)
    ClientData clientData;
    Display *display;		/* Not used. */
    char *widgRec;
    int offset;
{
    TreeViewIcon icon = *(TreeViewIcon *)(widgRec + offset);
    TreeView *tvPtr = clientData;

    Blt_TreeViewFreeIcon(tvPtr, icon);
}

static TreeViewStyleClass textEntryClass = {
    textEntrySpecs,
    ConfigTextEntry,
    DrawTextEntry,
    MeasureTextEntry,
    FreeTextEntry,
};

static TreeViewStyleClass checkBoxClass = {
    checkBoxSpecs,
    ConfigCheckBox,
    DrawCheckBox,
    MeasureCheckBox,
    FreeCheckBox,
};

static TreeViewStyleClass optionMenuClass = {
    optionMenuSpecs,
    ConfigOptionMenu,
    DrawOptionMenu,
    MeasureOptionMenu,
    FreeOptionMenu,
};

static TreeViewStyle *
CreateTextEntry(tvPtr)
    TreeView *tvPtr;
{
    TreeViewTextEntry *tePtr;

    tePtr = Blt_Calloc(1, sizeof(TreeViewTextEntry));
    assert(tePtr);
    tePtr->classPtr = &textEntryClass;
    tePtr->side = SIDE_LEFT;
    return (TreeViewStyle *)tePtr;
}

static void
ConfigTextEntry(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
    GC newGC;
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;
    XColor *bgColor;
    XGCValues gcValues;
    unsigned long gcMask;

    gcMask = GCForeground | GCBackground | GCFont;
    gcValues.font = Tk_FontId(CHOOSE(tvPtr->font, tePtr->font));
    bgColor = Tk_3DBorderColor(tvPtr->border);

    gcValues.background = CHOOSE(bgColor, tePtr->bgColor)->pixel;
    gcValues.foreground = CHOOSE(tvPtr->fgColor, tePtr->fgColor)->pixel;
    newGC = Tk_GetGC(tvPtr->tkwin, gcMask, &gcValues);
    if (tePtr->gc != NULL) {
	Tk_FreeGC(tvPtr->display, tePtr->gc);
    }
    tePtr->gc = newGC;
    gcValues.background = tePtr->highlightBgColor->pixel;
    gcValues.foreground = tePtr->highlightFgColor->pixel;
    newGC = Tk_GetGC(tvPtr->tkwin, gcMask, &gcValues);
    if (tePtr->highlightGC != NULL) {
	Tk_FreeGC(tvPtr->display, tePtr->highlightGC);
    }
    tePtr->highlightGC = newGC;

    gcValues.background = tePtr->activeBgColor->pixel;
    gcValues.foreground = tePtr->activeFgColor->pixel;
    newGC = Tk_GetGC(tvPtr->tkwin, gcMask, &gcValues);
    if (tePtr->activeGC != NULL) {
	Tk_FreeGC(tvPtr->display, tePtr->activeGC);
    }
    tePtr->activeGC = newGC;
    tePtr->flags |= STYLE_DIRTY;
}

static void
DrawTextEntry(tvPtr, drawable, entryPtr, valuePtr, stylePtr, x, y)
    TreeView *tvPtr;
    Drawable drawable;
    TreeViewEntry *entryPtr;
    TreeViewValue *valuePtr;
    TreeViewStyle *stylePtr;
    int x, y;
{
    GC gc;
    TreeViewColumn *columnPtr;
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;
    int iconWidth, iconHeight, textWidth, textHeight;
    int ix, iy, tx, ty;
    int gap;

    columnPtr = valuePtr->columnPtr;
    if (tePtr->flags & STYLE_HIGHLIGHT) {
	gc = tePtr->highlightGC;
    } else {
	gc = tePtr->gc;
    }
    if (!Blt_TreeViewEntryIsSelected(tvPtr, entryPtr)) {
	/*
	 * Draw the active or normal background color over the entire
	 * label area.  This includes both the tab's text and image.
	 * The rectangle should be 2 pixels wider/taller than this
	 * area. So if the label consists of just an image, we get an
	 * halo around the image when the tab is active.
	 */
#ifdef notdef
	XFillRectangle(tvPtr->display, drawable, gc, x, y, columnPtr->maxWidth,
		       entryPtr->height);
#endif
    }    
    tx = ty = ix = iy = 0;	/* Suppress compiler warning. */
    
    iconWidth = iconHeight = 0;
    if (tePtr->icon != NULL) {
	iconWidth = TreeViewIconWidth(tePtr->icon);
	iconHeight = TreeViewIconHeight(tePtr->icon);
    }
    textWidth = textHeight = 0;
    if (valuePtr->textPtr != NULL) {
	textWidth = valuePtr->textPtr->width;
	textHeight = valuePtr->textPtr->height;
    }
    gap = 0;
    if ((tePtr->icon != NULL) && (valuePtr->textPtr != NULL)) {
	gap = TEXT_ENTRY_GAP;
    }
    switch (tePtr->side) {
    case SIDE_RIGHT:
	tx = x;
	ty = y + (entryPtr->height - textHeight) / 2;
	ix = tx + textWidth + gap;
	iy = y + (entryPtr->height - iconHeight) / 2;
	break;
    case SIDE_LEFT:
	ix = x;
	iy = y + (entryPtr->height - iconHeight) / 2;
	tx = ix + iconWidth + gap;
	ty = y + (entryPtr->height - textHeight) / 2;
	break;
    case SIDE_TOP:
	iy = y;
	ix = x + (columnPtr->maxWidth - iconWidth) / 2;
	ty = iy + iconHeight + gap;
	tx = x + (columnPtr->maxWidth - textWidth) / 2;
	break;
    case SIDE_BOTTOM:
	ty = y;
	tx = x + (columnPtr->maxWidth - textWidth) / 2;
	iy = ty + textHeight + gap;
	ix = x + (columnPtr->maxWidth - iconWidth) / 2;
	break;
    }
    if (tePtr->icon != NULL) {
	Tk_RedrawImage(TreeViewIconBits(tePtr->icon), 0, 0, iconWidth, 
		       iconHeight, drawable, ix, iy);
    }
    if (valuePtr->textPtr != NULL) {
	TextStyle ts;
	XColor *color, *oldColor;
	Tk_Font font;

	oldColor = CHOOSE(tvPtr->fgColor, tePtr->fgColor);
	font = CHOOSE(tvPtr->font, tePtr->font);
	if (entryPtr->color != NULL) {
	    XSetForeground(tvPtr->display, gc, entryPtr->color->pixel);
	    color = entryPtr->color;
	} else {
	    color = oldColor;
	}
	Blt_SetDrawTextStyle(&ts, font, gc, color, oldColor, NULL, 0.0, 
		TK_ANCHOR_NW, TK_JUSTIFY_LEFT, 0, 0);
	Blt_DrawTextLayout(tvPtr->tkwin, drawable, valuePtr->textPtr, 
		&ts, tx, ty);
	if (color != oldColor) {
	    XSetForeground(tvPtr->display, gc, oldColor->pixel);
	}
    }
}

static void
FreeTextEntry(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;

    if (tePtr->highlightGC != NULL) {
	Tk_FreeGC(tvPtr->display, tePtr->highlightGC);
    }
    if (tePtr->activeGC != NULL) {
	Tk_FreeGC(tvPtr->display, tePtr->activeGC);
    }
    if (tePtr->gc != NULL) {
	Tk_FreeGC(tvPtr->display, tePtr->gc);
    }
    if (tePtr->icon != NULL) {
	Blt_TreeViewFreeIcon(tvPtr, tePtr->icon);
    }
}

static void
MeasureTextEntry(tvPtr, stylePtr, valuePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
    TreeViewValue *valuePtr;
{
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;
    int iconWidth, iconHeight;
    int textWidth, textHeight;
    int gap;

    textWidth = textHeight = 0;
    iconWidth = iconHeight = 0;
    valuePtr->width = valuePtr->height = 0;

    if (tePtr->icon != NULL) {
	iconWidth = TreeViewIconWidth(tePtr->icon);
	iconHeight = TreeViewIconHeight(tePtr->icon);
    } 
    if (valuePtr->textPtr != NULL) {
	Blt_Free(valuePtr->textPtr);
	valuePtr->textPtr = NULL;
    }
    if (valuePtr->string != NULL) {	/* New string defined. */
	TextStyle ts;

	Blt_InitTextStyle(&ts);
	ts.font = CHOOSE(tvPtr->font, tePtr->font);
	ts.anchor = TK_ANCHOR_NW;
	ts.justify = TK_JUSTIFY_LEFT;
	valuePtr->textPtr = Blt_GetTextLayout(valuePtr->string, &ts);
    } 
    gap = 0;
    if (valuePtr->textPtr != NULL) {
	textWidth = valuePtr->textPtr->width;
	textHeight = valuePtr->textPtr->height;
	if (tePtr->icon != NULL) {
	    gap = TEXT_ENTRY_GAP;
	}
    }
    if (SIDE_VERTICAL(tePtr->side)) {
	valuePtr->height = iconHeight + gap + textHeight;
	valuePtr->width = MAX(textWidth, iconWidth);
    } else {
	valuePtr->width = iconWidth + gap + textWidth;
	valuePtr->height = MAX(textHeight, iconHeight);
    }
}

static TreeViewStyle *
CreateCheckBox(tvPtr)
     TreeView *tvPtr;
{
    TreeViewCheckBox *checkBoxPtr;

    checkBoxPtr = Blt_Calloc(1, sizeof(TreeViewCheckBox));
    checkBoxPtr->classPtr = &checkBoxClass;
    return (TreeViewStyle *)checkBoxPtr;
}

static void
ConfigCheckBox(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
}

static void
DrawCheckBox(tvPtr, drawable, entryPtr, valuePtr, stylePtr, x, y)
    TreeView *tvPtr;
    Drawable drawable;
    TreeViewEntry *entryPtr;
    TreeViewValue *valuePtr;
    TreeViewStyle *stylePtr;
    int x, y;
{
}

static void
FreeCheckBox(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
}

static void
MeasureCheckBox(tvPtr, stylePtr, valuePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
    TreeViewValue *valuePtr;
{
}

static TreeViewStyle *
CreateOptionMenu(tvPtr)
     TreeView *tvPtr;
{
    TreeViewOptionMenu *optionMenuPtr;

    optionMenuPtr = Blt_Calloc(1, sizeof(TreeViewOptionMenu));
    optionMenuPtr->classPtr = &optionMenuClass;
    return (TreeViewStyle *)optionMenuPtr;
}

static void
ConfigOptionMenu(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
}

static void
DrawOptionMenu(tvPtr, drawable, entryPtr, valuePtr, stylePtr, x, y)
    TreeView *tvPtr;
    Drawable drawable;
    TreeViewEntry *entryPtr;
    TreeViewValue *valuePtr;
    TreeViewStyle *stylePtr;
    int x, y;
{
}

static void
FreeOptionMenu(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
}

static void
MeasureOptionMenu(tvPtr, stylePtr, valuePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
    TreeViewValue *valuePtr;
{
}

TreeViewStyle *
Blt_TreeViewNameToStyle(interp, tvPtr, styleName)
    Tcl_Interp *interp;
    TreeView *tvPtr;
    char *styleName;
{
    Blt_HashEntry *hPtr;

    hPtr = Blt_FindHashEntry(&tvPtr->styleTable, styleName);
    if (hPtr == NULL) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, "can't find cell style \"", styleName, 
		"\"", (char *)NULL);
	}
	return NULL;
    }
    return Blt_GetHashValue(hPtr);
}

int
Blt_TreeViewGetStyle(interp, tvPtr, styleName, stylePtrPtr)
    Tcl_Interp *interp;
    TreeView *tvPtr;
    char *styleName;
    TreeViewStyle **stylePtrPtr;
{
    TreeViewStyle *stylePtr;

    stylePtr = Blt_TreeViewNameToStyle(interp, tvPtr, styleName);
    if (stylePtr == NULL) {
	return TCL_ERROR;
    }
    stylePtr->refCount++;
    *stylePtrPtr = stylePtr;
    return TCL_OK;
}

TreeViewStyle *
Blt_TreeViewCreateStyle(interp, tvPtr, type, styleName)
     Tcl_Interp *interp;
     TreeView *tvPtr;		/* TreeView widget. */
     int type;			/* Type of style: either
				 * STYLE_TEXT_ENTRY,
				 * STYLE_OPTION_MENU, or
				 * STYLE_CHECK_BOX */
    char *styleName;		/* Name of the new style. */
{    
    Blt_HashEntry *hPtr;
    int isNew;
    TreeViewStyle *stylePtr;
    static char *classNames[3] = {
	"TextEntryStyle", "CheckBoxStyle", "OptionMenuStyle"
    };
    
    hPtr = Blt_CreateHashEntry(&tvPtr->styleTable, styleName, &isNew);
    if (!isNew) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, "cell style \"", styleName, 
			     "\" already exists", (char *)NULL);
	}
	return NULL;
    }
    /* Create the new marker based upon the given type */
    switch (type) {
    case STYLE_TEXT_ENTRY:
	stylePtr = CreateTextEntry(tvPtr);
	break;
    case STYLE_OPTION_MENU:
	stylePtr = CreateOptionMenu(tvPtr);
	break;
    case STYLE_CHECK_BOX:
	stylePtr = CreateCheckBox(tvPtr);
	break;
    default:
	return NULL;
    }
    assert(stylePtr);
    stylePtr->name = Blt_Strdup(styleName);
    stylePtr->flags = type;
    stylePtr->refCount = 1;
    stylePtr->hashPtr = hPtr;
    Blt_SetHashValue(hPtr, stylePtr);
    iconOption.clientData = tvPtr;
    if (Blt_ConfigureComponentFromObj(interp, tvPtr->tkwin, styleName, 
	classNames[type], stylePtr->classPtr->specsPtr, 0, (Tcl_Obj **)NULL, 
	(char *)stylePtr, 0) != TCL_OK) {
	Blt_TreeViewFreeStyle(tvPtr, stylePtr);
	return NULL;
    }
    (*stylePtr->classPtr->configProc)(tvPtr, stylePtr);
    return stylePtr;
}

void
Blt_TreeViewFreeStyle(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
    stylePtr->refCount--;
    /* Remove the style from the hash table so that it's name can be used.*/
    if (stylePtr->hashPtr != NULL) {
	Blt_DeleteHashEntry(&tvPtr->styleTable, stylePtr->hashPtr);
	stylePtr->hashPtr = NULL; 
    } 
    /* If no cell is using the style, remove it.*/
    if (stylePtr->refCount <= 0) {
	Blt_FreeObjOptions(stylePtr->classPtr->specsPtr, (char *)stylePtr, 
		   tvPtr->display, 0);
	(*stylePtr->classPtr->freeProc)(tvPtr, stylePtr); 
    } 
}

void
Blt_TreeViewSetTextEntryIcon(tvPtr, stylePtr, icon)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
    TreeViewIcon icon;
{
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;

    if (tePtr->icon != NULL) {
	Blt_TreeViewFreeIcon(tvPtr, tePtr->icon);
    }
    tePtr->icon = icon;
}

GC
Blt_TreeViewGetStyleGC(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;
    return tePtr->gc;
}

Tk_Font
Blt_TreeViewGetStyleFont(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;

    if (tePtr->font != NULL) {
	return tePtr->font;
    }
    return tvPtr->font;
}

XColor *
Blt_TreeViewGetStyleFg(tvPtr, stylePtr)
    TreeView *tvPtr;
    TreeViewStyle *stylePtr;
{
    TreeViewTextEntry *tePtr = (TreeViewTextEntry *)stylePtr;

    if (tePtr->fgColor != NULL) {
	return tePtr->fgColor;
    }
    return tvPtr->fgColor;
}

/*
 *----------------------------------------------------------------------
 *
 * StyleCgetOp --
 *
 *	  .t style cget "styleName" -background
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
StyleCgetOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;
    Tcl_Interp *interp;
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;
{
    TreeViewStyle *stylePtr;

    stylePtr = Blt_TreeViewNameToStyle(interp, tvPtr, Tcl_GetString(objv[3]));
    if (stylePtr == NULL) {
	return TCL_ERROR;
    }
    return Blt_ConfigureValueFromObj(interp, tvPtr->tkwin, 
	stylePtr->classPtr->specsPtr, (char *)tvPtr, objv[4], 0);
}

/*
 *----------------------------------------------------------------------
 *
 * StyleConfigureOp --
 *
 * 	This procedure is called to process a list of configuration
 *	options database, in order to reconfigure a style.
 *
 *	  .t style configure "styleName" option value
 *
 * Results:
 *	A standard Tcl result.  If TCL_ERROR is returned, then
 *	interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as text string, colors, font,
 *	etc. get set for stylePtr; old resources get freed, if there
 *	were any.  
 *
 *----------------------------------------------------------------------
 */
static int
StyleConfigureOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    TreeViewStyle *stylePtr;

    stylePtr = Blt_TreeViewNameToStyle(interp, tvPtr, Tcl_GetString(objv[3]));
    if (stylePtr == NULL) {
	return TCL_ERROR;
    }
    if (objc == 4) {
	return Blt_ConfigureInfoFromObj(interp, tvPtr->tkwin, 
	    stylePtr->classPtr->specsPtr, (char *)stylePtr, (Tcl_Obj *)NULL, 0);
    } else if (objc == 5) {
	return Blt_ConfigureInfoFromObj(interp, tvPtr->tkwin, 
		stylePtr->classPtr->specsPtr, (char *)stylePtr, objv[5], 0);
    }
    if (Blt_ConfigureWidgetFromObj(interp, tvPtr->tkwin, 
	stylePtr->classPtr->specsPtr, objc - 4, objv + 4, (char *)stylePtr, 
	BLT_CONFIG_OBJV_ONLY) != TCL_OK) {
	return TCL_ERROR;
    }
    Blt_TreeViewEventuallyRedraw(tvPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StyleCreateOp --
 *
 *	  .t style create type "styleName" -background blue
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
StyleCreateOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;
    Tcl_Interp *interp;
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;
{
    TreeViewStyle *stylePtr;
    int type;
    char *string;
    int isNew;
    Blt_HashEntry *hPtr;

    string = Tcl_GetString(objv[3]);
    if (strcmp(string, "text") == 0) {
	type = STYLE_TEXT;
    } else if (strcmp(string, "checkbox") == 0) {
	type = STYLE_CHECKBOX;
    } else if (strcmp(string, "optionmenu") == 0) {
	type = STYLE_OPTIONMENU;
    } else {
	Tcl_AppendResult(interp, "unknown style type \"", string, 
		"\": should be \"text\", \"checkbox\", or \"combobox\".", 
		(char *)NULL);
	return TCL_ERROR;
    }
    string = Tcl_GetString(objv[4]);
    hPtr = Blt_CreateHashEntry(&tvPtr->styleTable, string, &isNew);
    if (!isNew) {
	Tcl_AppendResult(interp, "style \"", string, "\" already exists.", 
		(char *)NULL);
	return TCL_ERROR;
    } 
    stylePtr = Blt_TreeViewCreateStyle(interp, tvPtr, type, string);
    if (stylePtr == NULL) {
	return TCL_ERROR;
    }
    stylePtr->hashPtr = hPtr;
    if (Blt_ConfigureWidgetFromObj(interp, tvPtr->tkwin, 
	stylePtr->classPtr->specsPtr, objc - 5, objv + 5, (char *)stylePtr, 
	0) != TCL_OK) {
	Blt_TreeViewFreeStyle(tvPtr, stylePtr);
	return TCL_ERROR;
    }
    Blt_SetHashValue(hPtr, stylePtr);
    Tcl_SetObjResult(interp, objv[4]);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StyleDeleteOp --
 *
 * 	Deletes one or more styles.  A style still may be used
 *	after it has been deleted if it's still in use.
 *
 *	  .t style delete "styleName"...
 *
 * Results:
 *	A standard Tcl result.  If TCL_ERROR is returned, then
 *	interp->result contains an error message.
 *
 *----------------------------------------------------------------------
 */
static int
StyleDeleteOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;

    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    TreeViewStyle *stylePtr;
    int i;

    for (i = 3; i < objc; i++) {
	stylePtr = Blt_TreeViewNameToStyle(interp, tvPtr, 
					   Tcl_GetString(objv[i]));
	if (stylePtr == NULL) {
	    return TCL_ERROR;
	}
	Blt_TreeViewFreeStyle(tvPtr, stylePtr);
    }
    Blt_TreeViewEventuallyRedraw(tvPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StyleNamesOp --
 *
 * 	Lists the names of all the current styles in the treeview widget.
 *
 *	  .t style names
 *
 * Results:
 *	Always TCL_OK.
 *
 *----------------------------------------------------------------------
 */
static int
StyleNamesOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;

    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    Blt_HashEntry *hPtr;
    Blt_HashSearch cursor;
    Tcl_Obj *listObjPtr, *objPtr;
    TreeViewStyle *stylePtr;

    listObjPtr = Tcl_NewListObj(0, (Tcl_Obj **)NULL);
    for (hPtr = Blt_FirstHashEntry(&tvPtr->styleTable, &cursor); hPtr != NULL;
	 hPtr = Blt_NextHashEntry(&cursor)) {
	stylePtr = Blt_GetHashValue(hPtr);
	objPtr = Tcl_NewStringObj(stylePtr->name, -1);
	Tcl_ListObjAppendElement(interp, listObjPtr, objPtr);
    }
    Tcl_SetObjResult(interp, listObjPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StyleSetOp --
 *
 * 	Sets a style for a given key for all the ids given.
 *
 *	  .t style set styleName key node...
 *
 * Results:
 *	A standard Tcl result.  If TCL_ERROR is returned, then
 *	interp->result contains an error message.
 *
 *----------------------------------------------------------------------
 */
static int
StyleSetOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;

    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    Blt_TreeKey key;
    TreeViewEntry *entryPtr;
    TreeViewStyle *stylePtr, *oldStylePtr;
    TreeViewTagInfo info;
    int i;

    stylePtr = Blt_TreeViewNameToStyle(interp, tvPtr, Tcl_GetString(objv[3]));
    if (stylePtr == NULL) {
	return TCL_ERROR;
    }
    key = Blt_TreeGetKey(Tcl_GetString(objv[4]));
    stylePtr->flags |= STYLE_LAYOUT;
    for (i = 5; i < objc; i++) {
	if (Blt_TreeViewFindTaggedEntries(tvPtr, objv[i], &info) != TCL_OK) {
	    return TCL_ERROR;
	}
	for (entryPtr = Blt_TreeViewFirstTaggedEntry(&info); entryPtr != NULL; 
	     entryPtr = Blt_TreeViewNextTaggedEntry(&info)) {
	    register TreeViewValue *valuePtr;

	    for (valuePtr = entryPtr->values; valuePtr != NULL; 
		 valuePtr = valuePtr->nextPtr) {
		if (valuePtr->columnPtr->key == key) {
		    stylePtr->refCount++;
		    oldStylePtr = valuePtr->stylePtr;
		    valuePtr->stylePtr = stylePtr;
		    if (oldStylePtr != NULL) {
			Blt_TreeViewFreeStyle(tvPtr, oldStylePtr);
		    }
		    break;
		}
	    }
	}
    }
    Blt_TreeViewEventuallyRedraw(tvPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StyleUnsetOp --
 *
 * 	Removes a style for a given key for all the ids given.
 *	The cell's style is returned to its default state.
 *
 *	  .t style unset styleName key node...
 *
 * Results:
 *	A standard Tcl result.  If TCL_ERROR is returned, then
 *	interp->result contains an error message.
 *
 *----------------------------------------------------------------------
 */
static int
StyleUnsetOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;

    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    Blt_TreeKey key;
    TreeViewEntry *entryPtr;
    TreeViewStyle *stylePtr;
    TreeViewTagInfo info;
    int i;

    stylePtr = Blt_TreeViewNameToStyle(interp, tvPtr, Tcl_GetString(objv[3]));
    if (stylePtr == NULL) {
	return TCL_ERROR;
    }
    key = Blt_TreeGetKey(Tcl_GetString(objv[4]));
    stylePtr->flags |= STYLE_LAYOUT;
    for (i = 5; i < objc; i++) {
	if (Blt_TreeViewFindTaggedEntries(tvPtr, objv[i], &info) != TCL_OK) {
	    return TCL_ERROR;
	}
	for (entryPtr = Blt_TreeViewFirstTaggedEntry(&info); entryPtr != NULL; 
	     entryPtr = Blt_TreeViewNextTaggedEntry(&info)) {
	    register TreeViewValue *valuePtr;

	    for (valuePtr = entryPtr->values; valuePtr != NULL; 
		 valuePtr = valuePtr->nextPtr) {
		if (valuePtr->columnPtr->key == key) {
		    if (valuePtr->stylePtr != NULL) {
			Blt_TreeViewFreeStyle(tvPtr, valuePtr->stylePtr);
			valuePtr->stylePtr = NULL;
		    }
		    break;
		}
	    }
	}
    }
    Blt_TreeViewEventuallyRedraw(tvPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StyleOp --
 *
 *	.t style cget "highlight" -foreground
 *	.t style configure "highlight" -fg blue -bg green
 *	.t style create checkbox "highlight"
 *	.t style create optionmenu "highlight"
 *	.t style create text "highlight"
 *	.t style delete "highlight"
 *	.t style get "mtime" $node
 *	.t style names
 *	.t style set "mtime" "highlight" all
 *	.t style unset "mtime" all
 *
 *---------------------------------------------------------------------- 
 */
static Blt_OpSpec styleOps[] = {
    {"cget", 2, (Blt_Op)StyleCgetOp, 5, 5, "styleName option",},
    {"configure", 2, (Blt_Op)StyleConfigureOp, 4, 0, "styleName options...",},
    {"create", 1, (Blt_Op)StyleCreateOp, 5, 0, "type styleName options...",},
    {"delete", 2, (Blt_Op)StyleDeleteOp, 3, 0, "styleName...",},
    {"names", 1, (Blt_Op)StyleNamesOp, 3, 3, "",}, 
    {"set", 1, (Blt_Op)StyleSetOp, 6, 6, "key styleName tagOrId...",},
    {"unset", 1, (Blt_Op)StyleUnsetOp, 5, 5, "key tagOrId",},
};

static int nStyleOps = sizeof(styleOps) / sizeof(Blt_OpSpec);

int
Blt_TreeViewStyleOp(tvPtr, interp, objc, objv)
    TreeView *tvPtr;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    Blt_Op proc;
    int result;

    proc = Blt_GetOpFromObj(interp, nStyleOps, styleOps, BLT_OP_ARG2, objc, 
	objv, 0);
    if (proc == NULL) {
	return TCL_ERROR;
    }
    result = (*proc)(tvPtr, interp, objc, objv);
    return result;
}

#endif /* NO_TREEVIEW */


